module Main ( main
            ) where

import Algos.Graphs.Graph (Graph(..), Vertex(..), Edge(..))
import Algos.Graphs.GraphIO (Display(..), displayGraph, Orientation(..))
import Algos.Random.Numbers (randomNumberM)
import Algos.Random.Sequences (randomPermutationM, randomSubPermutationM)
import Algos.Random.Sets (randomCombinationM)
import Control.Applicative ((<$>))
import Control.Exception (SomeException(..), handle)
import Control.Monad.State (State, evalState)
import Data.Array (Array, listArray, (!))
import Data.List (sort)
import Data.Map (fromAscList, lookupLT)
import System.Environment (getArgs, getProgName)
import System.Exit (exitFailure)
import System.IO (IOMode(..), hPutStr, withFile)
import System.Random (RandomGen, newStdGen)

main :: IO ()
main =  do (n, m, b, file) <- readArgs
           gen             <- newStdGen
           let g = evalState (genGraph b n m) gen
           withFile file WriteMode $
             \h -> hPutStr h (displayGraph (Undirected, g))

readArgs :: IO (Int, Int, Bool, String)
readArgs =  handle (\(SomeException _) -> usage >> exitFailure) $
            do [nstr, mstr, bstr, file] <- getArgs
               let n = read nstr
                   m = read mstr
                   b = read bstr
               return (n, m, b, file)

usage :: IO ()
usage =  do prog <- getProgName
            putStrLn $ "USAGE: " ++ prog ++ " <n> <m> <True|False> <graph file>"

genGraph       :: RandomGen g => Bool -> Int -> Int -> State g (Graph Int Int Int)
genGraph True  =  genBipartite
genGraph False =  genWithOddCycle

genBipartite     :: RandomGen g => Int -> Int -> State g (Graph Int Int Int)
genBipartite n m =  do vs <- randomPermutationM [Vertex i 0 | i <- [1..n]]
                       s  <- randomNumberM (1,n-1)
                       let (us, ws) = splitAt s vs
                           [nu, nw] = map length [us, ws]
                           mmax     = nu * nw
                           m'       = min mmax m
                           ua       = listArray (1, nu) us
                           wa       = listArray (1, nw) ws
                       ixs <- randomSubPermutationM m' mmax
                       es  <- randomPermutationM (zipWith (bipIxToEdge nu nw ua wa) ixs [1..])
                       return (Graph vs es)

bipIxToEdge                  :: Int -> Int -> Array Int (Vertex Int Int) -> Array Int (Vertex Int Int) -> 
                                Int -> Int -> Edge Int Int
bipIxToEdge nu nw ua wa ix i =  Edge u w i
  where wix = (ix-1) `div` nu
        uix = (ix-1) - wix * nu
        u   = vertexId (ua ! (uix + 1))
        w   = vertexId (wa ! (wix + 1))

genWithOddCycle     :: RandomGen g => Int -> Int -> State g (Graph Int Int Int)
genWithOddCycle n m =  do let ncmax' = minimum [n, m]
                              ncmax | odd ncmax' = ncmax' + 1
                                    | otherwise  = ncmax'
                          nc' <- randomNumberM (3, ncmax)
                          let nc | odd nc'   = nc'
                                 | otherwise = nc' - 1
                          let ces  = (1,nc) : [(i,i+1) | i <- [1..nc-1]]
                              cixs = map edgeToIx ces
                              mmax = (n * (n-1) `div` 2) - nc
                              m'   = min mmax (m - nc)
                          ixs  <- adjustIndices cixs <$> randomCombinationM m' mmax
                          let nces = ixsToEdges ixs
                          vids <- randomPermutationM [1..n]
                          eids <- randomPermutationM [1..m' + nc]
                          let va = listArray (1,n) vids
                          vs <- randomPermutationM (map (flip Vertex 0) vids)
                          es <- randomPermutationM (zipWith (mkEdge va) (ces ++ nces) eids)
                          return (Graph vs es)

edgeToIx        :: (Int, Int) -> Int
edgeToIx (u, v) |  v < u     = edgeToIx (v, u)
                |  otherwise = v * (v - 1) + u + 1

adjustIndices           :: [Int] -> [Int] -> [Int]
adjustIndices exixs ixs =  map adjIx ixs
  where exmap    = fromAscList $ zipWith (\ix adj -> (ix-adj, adj)) (sort exixs) [1..]
        adjIx ix = ix + maybe 0 snd (lookupLT ix exmap)
        

ixsToEdges :: [Int] -> [(Int, Int)]
ixsToEdges =  go 2 0
  where go _ _   []  = []
        go v off ixs = map (\u -> (u-off, v)) ixs' ++ go (v+1) off' ixs''
          where off'          = off + v - 1
                (ixs', ixs'') = span (<= off') ixs

mkEdge             :: Array Int Int -> (Int, Int) -> Int -> Edge Int Int
mkEdge va (u, v) i =  Edge (va ! u) (va ! v) i

instance Display Int where
  display = show
