{-------------------------------------------------------------------------------

A wrapper for the bipartiteness testing algorithm.  The input graph is read from
file, which must have the following structure:

- One line storing the word "Directed" or "Undirected" and two integers, one the
  number n of vertices, the other the number m of edges.  (This is a consequence
  of using Algos.GraphIO.parseGraph to read the graph.  This labelling is ignored
  by the algorithm.)

- n lines, one per vertex, storing the ID and label of each vertex separated by
  spaces.

- m lines, one per edge, storing the ID of the tail, the ID of the head, and the
  label separated by spaces.

This code makes no effort to check that the input is actually valid.  For
example, if fewer than m edge lines are created, the code will simply raise an
exception and terminate abnormally.

The output is represented as a single line containing the word True or False
to indicate whether the graph is bipartite or not.  If the graph is bipartite,
this is followed by two lines containing the list of vertices in each side of
the bipartition, respectively.  If the graph is not bipartite, a single line
lists the vertices of an odd cycle in the order they occur on this cycle.

-------------------------------------------------------------------------------}

{-# LANGUAGE FlexibleInstances #-}

module Main ( main
            ) where

import Algos.Graphs.Bipartiteness (isBipartite)
import Algos.Graphs.Graph (Graph, AdjList(..), makeUndirAdjListIx, E(..))
import Algos.Graphs.GraphIO (Display(..), Parse(..), parseGraph)
import Control.Applicative ((<$>))
import Data.List (intercalate)
import System.Environment (getArgs, getProgName)
import System.Exit (exitFailure)
import System.IO (withFile, IOMode(..), hPutStrLn)

-- The main function
main :: IO ()
main =  do (gfile, bipfile) <- readArgs
           (_, g) <- parseGraph <$> readFile gfile 
           writeBip bipfile (runBip g)

readArgs :: IO (String, String)
readArgs =  do args <- getArgs
               case args of
                 [gfile, bipfile] -> return (gfile, bipfile)
                 _                -> usage >> exitFailure

-- Prints a usage message in case the command line arguments make no sense
usage :: IO ()
usage =  do prog <- getProgName
            putStrLn $ "USAGE: " ++ prog ++ " <input file> <output file>"
           
-- Prints a given graph to stdout
writeBip           :: String -> Either [Int] ([Int], [Int]) -> IO ()
writeBip fname bip =  withFile fname WriteMode $
                      \h -> hPutStrLn h $ 
                            either (\vs       -> "False\n" ++ intercalate " " (map show vs))
                                   (\(us, ws) -> "True\n" ++ intercalate " " (map show us) ++
                                                 "\n" ++ intercalate " " (map show ws))
                                   bip

-- Tests whether the graph is bipartite
runBip   :: Graph Int Int Int -> Either [Int] ([Int], [Int])
runBip g =  case isBipartite a of
              Left  es       -> Left  (map (adjVertexId a . eTail) es)
              Right (us, ws) -> Right (map (adjVertexId a) us, map (adjVertexId a) ws)
  where a = makeUndirAdjListIx g

-- A parse instance for strings that doesn't require quotation marks
instance Parse Int where
  parse = read

-- A display instance for strings that doesn't print quotation marks
instance Display Int where
  display = show
