module Main ( main
            ) where

import Algos.Graphs.Graph (Graph(..), Edge(..))
import Algos.Graphs.GraphIO (Parse(..), parseGraph)
import Control.Applicative ((<$>))
import Data.Array (array, (!))
import Data.List (find)
import Data.Monoid
import Data.Set (fromList, member)
import System.Environment (getArgs, getProgName)
import System.Exit (exitFailure)

main :: IO ()
main =  do (gfile, bipfile) <- readArgs
           (_, g) <- parseGraph <$> readFile gfile
           bip    <- parseBip   <$> readFile bipfile
           putStrLn (formatResult $ verifyBip bip g)

readArgs :: IO (String, String)
readArgs =  do args <- getArgs
               case args of
                 [gfile, bipfile] -> return (gfile, bipfile)
                 _                -> usage >> exitFailure

usage :: IO ()
usage =  do prog <- getProgName
            putStrLn $ "USAGE: " ++ prog ++ " <graph file> <result file>"

parseBip     :: String -> Either [Int] ([Int], [Int])
parseBip txt |  isBip     = Right (us, ws)
             |  otherwise = Left  cvs
  where ls       = lines txt
        nums     = map (map read . words) (tail ls)
        isBip    = read (head ls)
        [us, ws] = nums
        [cvs]    = nums

formatResult                                      :: Result -> String
formatResult OK                                   =  "OK"
formatResult (NotAnEdge e)                        =  "Edge in cycle is not in graph: " ++ show e
formatResult (NotBipartite e)                     =  "Found edge with endpoints in same bipartition: " ++ show e

data Result = OK
            | NotAnEdge    (Int, Int)
            | NotBipartite (Int, Int)

instance Monoid Result where
  mempty       = OK
  mappend OK x = x
  mappend x  _ = x

verifyBip              :: Either [Int] ([Int], [Int]) -> Graph Int Int Int -> Result
verifyBip (Left cs)    =  verifyCycle cs
verifyBip (Right part) =  verifyBipartition part

verifyCycle      :: [Int] -> Graph Int Int Int -> Result
verifyCycle cs g =  maybe OK NotAnEdge e
  where es  = fromList (concatMap (\(Edge u v _) -> [(u, v), (v, u)]) (graphEdges g))
        ces = zip cs (tail $ cycle cs)
        e   = find (not . flip member es) ces

verifyBipartition            :: ([Int], [Int]) -> Graph Int Int Int -> Result
verifyBipartition (us, ws) g =  maybe OK (\(Edge u v _) -> NotBipartite (u, v)) e
  where n   = length us + length ws
        bip = array (1, n) $ zip us (repeat U) ++ zip ws (repeat W)
        e   = find (\(Edge u v _) -> bip ! u == bip ! v) (graphEdges g)

data Bip = U | W deriving Eq

instance Parse Int where
  parse = read
