{-------------------------------------------------------------------------------

A simple module that provides I/O functions for reading and writing graphs.  For
now, all we need is a function to read a graph and functions to print trees,
forests, and graphs.

-------------------------------------------------------------------------------}

module Algos.Graphs.GraphIO ( Display(..)
                            , Parse(..)
                            , parseGraph
                            , displayGraph
                            , Orientation(..)
                            , NewickType(..)
                            ) where

import Algos.Graphs.Forest(Forest(..),Tree(..),Subtree(..))
import Algos.Graphs.Graph(Graph(..),Vertex(..),Edge(..))
import Data.Char (isSpace)
import Data.List (intercalate, unfoldr)

-- Type class for parsable vertex and edge labels.  Needed to avoid the
-- requirement of quotation marks around string labels for parsing strings.
class Parse t where
  parse :: String -> t

-- Type class for printable vertex and edge labels.  Needed to avoid printing
-- quotation marks around string labels.
class Display t where
  display :: t -> String

-- Is the graph directed or undirected?
data Orientation = Directed | Undirected deriving (Eq, Read, Show)

-- Function to parse a string into a graph
--
-- The input graph is read from file, which must have the following structure:
--
-- - One line storing the word "Directed" or "Undirected" and two integers, one the
--   number n of vertices, the other the number m of edges.
--
-- - n lines, one per vertex, storing the ID and label of each vertex separated by
--   spaces.
--
-- - m lines, one per edge, storing the ID of the tail, the ID of the head, and the
--   label separated by spaces.
parseGraph   :: (Parse v, Parse vl, Parse el) => String -> (Orientation, Graph v vl el)
parseGraph s =  (o, Graph vs es)
  where  (hdr:ls)     = filter (not . null) (lines s)
         [os, ns, ms] = words hdr
         o            = read os
         [n, m]       = map read [ns, ms]
         (vss, ess)   = splitAt n ls
         vs           = map parseVertex vss
         es           = map parseEdge ess

         parseVertex vs = Vertex i l
           where [is, ls] = words vs
                 i        = parse is
                 l        = parse ls

         parseEdge es = Edge t h l
           where [ts, hs, ls] = words es
                 t            = parse ts
                 h            = parse hs
                 l            = parse ls

-- Function to display a graph.  Outputs the graph in exactly the format that
-- parseGraph expects.
displayGraph                  :: (Display v, Display vl, Display el) => (Orientation, Graph v vl el) -> String
displayGraph (o, Graph vs es) =  unlines (unwords [ostr, nstr, mstr] : vstrs ++ estrs)
  where ostr  = show o
        nstr  = show (length vs)
        mstr  = show (length es)
        vstrs = map displayVertex vs
        estrs = map displayEdge   es

        displayVertex (Vertex i l) = intercalate " " [display i, display l]
        displayEdge   (Edge t h l) = intercalate " " [display t, display h, display l]

-- Functions to parse and display a forest or tree in Newick format:
--
-- - A forest is represented as a sequence of trees terminated by a semicolon and
--   separated by newlines.
--
-- - A tree is represented as
--   - an opening parenthesis,
--   - the ID of the root,
--   - a colon if the root has at least one child,
--   - for each subtree, the label of the parent edge followed by the Newick string
--     for this subtree, and
--   - a closing parenthesis.        
class NewickType t where
  parseNewick   :: (Parse v, Parse e)     => String -> t v e
  displayNewick :: (Display v, Display e) => t v e -> String

instance NewickType Forest where
  parseNewick   = parseNewickForest
  displayNewick = displayNewickForest

instance NewickType Tree where
  parseNewick     = (\(Forest (t:_)) -> t) . parseNewickForest
  displayNewick t = displayNewickForest (Forest [t])

displayNewickForest             :: (Display v, Display e) => Forest v e -> String
displayNewickForest (Forest ts) =  tail $ foldr (displayNewickTree "\n") ";\n" ts

displayNewickTree                      :: (Display v, Display e) => String -> Tree v e -> String -> String
displayNewickTree join (Tree v ws) txt =  join ++ "(" ++ display v ++ addColon txt'
  where addColon xs@(')':_) =     xs
        addColon xs         = ':':xs
        txt'                = foldr displayNewickSubtree (')' : txt) ws

displayNewickSubtree                   :: (Display v, Display e) => Subtree v e -> String -> String
displayNewickSubtree (Subtree e t) txt =  display e ++ displayNewickTree "" t txt

parseNewickForest     :: (Parse v, Parse e) => String -> Forest v e
parseNewickForest txt =  Forest (unfoldr parseMaybeNewickTree txt)

parseMaybeNewickTree     :: (Parse v, Parse e) => String -> Maybe (Tree v e, String)
parseMaybeNewickTree txt | x == ';'  = Nothing
                         | otherwise = Just (parseNewickTree txt')
  where txt'@(x:_) = dropWhile isSpace txt

parseNewickTree           :: (Parse v, Parse e) => String -> (Tree v e, String)
parseNewickTree ('(':txt) =  (Tree v ws, txt'')
  where (vstr, txt'@(x:xs))     = break (`elem` [')', ':']) txt
        v                       = parse vstr
        (ws, txt'') | x == ':'  = unfoldr' parseMaybeNewickSubtree xs
                    | otherwise = ([], xs)

parseMaybeNewickSubtree     :: (Parse v, Parse e) => String -> (Maybe (Subtree v e), String)
parseMaybeNewickSubtree txt |  x == ')'  = (Nothing, xs)
                            |  otherwise = (Just (Subtree e t), txt''')
  where txt'@(x:xs)   = dropWhile isSpace txt
        (estr, txt'') = break (== '(') txt'
        e             = parse estr
        (t, txt''')   = parseNewickTree txt''

unfoldr'     :: (b -> (Maybe a, b)) -> b -> ([a], b)
unfoldr' f x =  unf x
  where unf x = maybe ([], x') g y
          where (y, x') = f x
                g y     = (y : ys, x'')
                  where (ys, x'') = unf x'
